import { copyFile, readFile } from 'fs/promises'
import inquirer from 'inquirer'
import { parseADOFAILevel } from './utils/parser.js'
import {
  ActionType,
  DecorationEvent,
  LevelEvent,
  MoveDecorations,
} from './types/events.js'
import { join } from 'path'
import { dirname } from 'path'
import sharp from 'sharp'
import { Level } from './types/level.js'

const { filePath } = await inquirer.prompt([
  {
    message: 'wow give me level file path',
    name: 'filePath',
  },
])
const levelDir = dirname(filePath)

const distPath = join(process.cwd(), 'dist')

console.log(filePath)

// todo functions
// .filter(x=>x.tag.split(' ').includes(tag))
// function findAndFilterDecorationEventWithTagReference(levelEvents: LevelEvent[], tag: string): LevelEvent[]

// occupiedTags: string[]
// let newTag = createUniqueTag(tag)
// findAndFilterDecorationEventWithTagReference(levelEvents, tag).forEachasfhgjksguhds
// function createTagAndCopyUsageReference(levelEvents: LevelEvent[], tag: string): LevelEvent[]

// function splitTagByCondition(levelEvents: LevelEvent[], tag: string, condition: (levelEvent: LevelEvent) => boolean): LevelEvent[]

// function minimizeEvent(levelEvents: LevelEvent[]): LevelEvent[]

class EventSans {
  origEvents: MoveDecorations[] = []
  events: MoveDecorations[] = []

  // decorationsByTag: Map<string, Decoration[]> = new Map()

  get decorations() {
    return this._level.decorations as Decoration[]
  }

  constructor(private _level: Level) {
    this.filterEvents()
  }

  filterEvents() {
    const actions = this._level.actions

    for (let index = actions.length - 1; index >= 0; index--) {
      const event = actions[index]

      if (event.eventType === ActionType.MoveDecorations) {
        actions.splice(index, 1)
        for (const tag of event.tag.split(' ')) {
          if (!tag) {
            this.origEvents.push({ ...event })
            continue
          }
          this.origEvents.push({ ...event, tag })
        }
      }
    }
  }

  /**
   * wow
   */
  tagDecorations() {
    // ? 뭐해야하지
    // of -> array의 엘리먼트 / 오브젝트의 value 부분
    // in -> array의 index를 string으로 만들어버린거 / 오브젝트의 key 부분

    let decorationsByTag: Map<string, Decoration[]> = new Map()

    const setOrPush = (tag: string, deco: Decoration) => {
      if (decorationsByTag.has(tag)) {
        let arr = decorationsByTag.get(tag)!.slice()
        arr.push(deco)

        decorationsByTag.set(tag, arr)
      } else decorationsByTag.set(tag, [deco])
    }

    // 이거 끝난거 아닌가(????) -> 원래 태그는 근데 가지고 있어야 나중에 육안으로 잘 됐는지 구분 가능할듯
    for (const deco of this.decorations) {
      const tag = `${deco.tag}-${deco.id}`

      for (const ev of this.origEvents) {
        if (ev.tag.split(' ').includes(deco.tag)) {
          // 이벤트에는 태그 여러개들어감..................... 스페이스로 구분해서............
          this.events.push({ ...ev, tag })
        }
      }

      deco.tag = tag
    }
  }
}

type Decoration = DecorationEvent & { id: number }
type ImageSize = { width: number; height: number }

// Read file
const content: string = await readFile(filePath, 'utf-8')

// Objectify content JSON.parse (or similar method)
const level = parseADOFAILevel(content)

const decorations: Decoration[] = []
const imagePaths: Map<string, ImageSize> = new Map()

let lastId = 0

const registerDecorationImage = async (decorationImage: string) => {
  if (!decorationImage || imagePaths.has(decorationImage)) return

  const imagePath = join(levelDir, decorationImage)

  const meta = await sharp(imagePath).metadata()

  imagePaths.set(decorationImage, {
    width: meta.width!,
    height: meta.height!,
  })
}

for (const event of [...level.actions, ...level.decorations]) {
  // Read through decorations, apply id to each decoration
  if (
    event.eventType === ActionType.AddDecoration ||
    event.eventType === ActionType.AddText
  ) {
    const decoration: Decoration = event as Decoration
    decoration.id = lastId++
    decorations.push(decoration)

    // Record file path in decorationImage property then acquire image resolution, match these
    if (decoration.eventType != ActionType.AddDecoration) continue

    await registerDecorationImage(decoration.decorationImage)
  }

  // Record file path in decorationImage property then acquire image resolution, match these
  if (event.eventType === ActionType.MoveDecorations)
    await registerDecorationImage(event.decorationImage)
}

// Create a new directory, then copy over all files to new directory
for (const [imagePath, image] of imagePaths) {
  let newPath = join(distPath, imagePath)

  // Apply resolution changes to image files

  await sharp(join(levelDir, imagePath))
    .resize({ width: image.width, height: image.height })
    .toFile(newPath)

  // 충격: 아직 파일별로 ratio로 나누는거 구현안함
  // wow
}

// Iterate through decorations:
// 1) record which tags are used in each decoration
// 2) split image decoration(AddDecoration eventType) and text decoration(AddText eventType) tag references
// 3) make changes to pivotOffset and scale properties depending on which image path is stored in decorationImage
//!4) ensure one to one relation to image files and image decoration tag by duplicating tag and its usage. duplicate the event if needed
// Iterate through actions
// 1) make changes to pivotOffset and scale properties corresponding to tag
// Save file to new directory
