import {
  SpeedType,
  BgDisplayMode,
  ToggleBool,
  TrackColorType,
  TrackColorPulse,
  TrackStyle,
  TrackAnimationType,
  TrackAnimationType2,
  DecPlacementType,
  DecorationBlendMode,
  MaskingType,
  Hitbox,
  FlashPlane,
  Ease,
  RandomMode,
  CamMovementType,
  GameSound,
  HitSound,
  Filter,
  EasePartBehavior,
  HoldStartSound,
  HoldLoopSound,
  HoldEndSound,
  HoldMidSound,
  HoldMidSoundType,
  HoldMidSoundTimingRelativeTo,
  PlanetCount,
  FontName,
  TargetPlanet,
} from './types'

export enum ActionType {
  SetSpeed = 'SetSpeed',
  Twirl = 'Twirl',
  Multitap = 'Multitap',
  Checkpoint = 'Checkpoint',
  CustomBackground = 'CustomBackground',
  ChangeTrack = 'ChangeTrack',
  ColorTrack = 'ColorTrack',
  AnimateTrack = 'AnimateTrack',
  AddDecoration = 'AddDecoration',
  Flash = 'Flash',
  MoveCamera = 'MoveCamera',
  SetHitsound = 'SetHitsound',
  SetGameSound = 'SetGameSound',
  PlaySound = 'PlaySound',
  RecolorTrack = 'RecolorTrack',
  MoveTrack = 'MoveTrack',
  SetFilter = 'SetFilter',
  HallOfMirrors = 'HallOfMirrors',
  ShakeScreen = 'ShakeScreen',
  SetPlanetRotation = 'SetPlanetRotation',
  KillPlayer = 'KillPlayer',
  MoveDecorations = 'MoveDecorations',
  PositionTrack = 'PositionTrack',
  Hold = 'Hold',
  SetHoldSound = 'SetHoldSound',
  TileDimensions = 'TileDimensions',
  RepeatEvents = 'RepeatEvents',
  Bloom = 'Bloom',
  SetConditionalEvents = 'SetConditionalEvents',
  MultiPlanet = 'MultiPlanet',
  FreeRoam = 'FreeRoam',
  FreeRoamTwirl = 'FreeRoamTwirl',
  FreeRoamRemove = 'FreeRoamRemove',
  FreeRoamWarning = 'FreeRoamWarning',
  Pause = 'Pause',
  AutoPlayTiles = 'AutoPlayTiles',
  Hide = 'Hide',
  ScaleMargin = 'ScaleMargin',
  ScaleRadius = 'ScaleRadius',
  ScreenTile = 'ScreenTile',
  ScreenScroll = 'ScreenScroll',
  AddText = 'AddText',
  SetText = 'SetText',
  EditorComment = 'EditorComment',
  Bookmark = 'Bookmark',
  CallMethod = 'CallMethod',
  AddComponent = 'AddComponent',
  ScalePlanets = 'ScalePlanets',
}

export interface Action<T extends ActionType> {
  floor: number
  eventType: T
}

export type Tile = [number, TileType]

export enum TileType {
  ThisTile = 'ThisTile',
  FirstTile = 'FirstTile',
  LastTile = 'LastTile',
}

export type Color = `#${string}`

export type Vector2 = [number, number]

export interface SetSpeed extends Action<ActionType.SetSpeed> {
  speedType: SpeedType
  beatsPerMinute: number
  bpmMultiplier: number
}

export interface Twirl extends Action<ActionType.Twirl> {}

export interface Multitap extends Action<ActionType.Multitap> {
  taps: number
}

export interface Checkpoint extends Action<ActionType.Checkpoint> {
  tileOffset: number
}

export interface CustomBackground extends Action<ActionType.CustomBackground> {
  color: Color
  bgImage: string
  imageColor: Color
  parallax: Vector2
  bgDisplayMode: BgDisplayMode
  lockRot: ToggleBool
  loopBG: ToggleBool
  unscaledSize: number
  angleOffset: number
  eventTag: string
}

export interface ChangeTrack extends Action<ActionType.ChangeTrack> {
  trackColorType: TrackColorType
  trackColor: Color
  secondaryTrackColor: Color
  trackColorAnimDuration: number
  trackColorPulse: TrackColorPulse
  trackPulseLength: number
  trackStyle: TrackStyle
  trackAnimation: TrackAnimationType
  beatsAhead: number
  trackDisappearAnimation: TrackAnimationType2
  beatsBehind: number
}

export interface ColorTrack extends Action<ActionType.ColorTrack> {
  trackColorType: TrackColorType
  trackColor: Color
  secondaryTrackColor: Color
  trackColorAnimDuration: number
  trackColorPulse: TrackColorPulse
  trackPulseLength: number
  trackStyle: TrackStyle
  trackTexture: string
  trackTextureScale: number
  trackGlowIntensity: number
}

export interface AnimateTrack extends Action<ActionType.AnimateTrack> {
  trackAnimation: TrackAnimationType
  beatsAhead: number
  trackDisappearAnimation: TrackAnimationType2
  beatsBehind: number
}

export interface AddDecoration extends Action<ActionType.AddDecoration> {
  decorationImage: string
  position: Vector2
  relativeTo: DecPlacementType
  pivotOffset: Vector2
  rotation: number
  lockRotation: ToggleBool
  scale: Vector2
  lockScale: ToggleBool
  tile: Vector2
  color: Color
  opacity: number
  depth: number
  parallax: Vector2
  parallaxOffset: Vector2
  tag: string
  imageSmoothing: ToggleBool
  blendMode: DecorationBlendMode
  maskingType: MaskingType
  useMaskingDepth: ToggleBool
  maskingFrontDepth: number
  maskingBackDepth: number
  failHitbox: ToggleBool
  failHitboxType: Hitbox
  failHitboxScale: Vector2
  failHitboxOffset: Vector2
  failHitboxRotation: number
  components: string
}

export interface Flash extends Action<ActionType.Flash> {
  duration: number
  plane: FlashPlane
  startColor: Color
  startOpacity: number
  endColor: Color
  endOpacity: number
  angleOffset: number
  ease: Ease
  eventTag: string
}

export interface MoveCamera extends Action<ActionType.MoveCamera> {
  duration: number
  durationRandomMode: RandomMode
  durationRandomValue: number
  relativeTo: CamMovementType
  position: Vector2
  positionRandomMode: RandomMode
  positionRandomValue: Vector2
  rotation: number
  rotationRandomMode: RandomMode
  rotationRandomValue: number
  zoom: number
  zoomRandomMode: RandomMode
  zoomRandomValue: number
  angleOffset: number
  ease: Ease
  dontDisable: ToggleBool
  minVfxOnly: ToggleBool
  eventTag: string
}

export interface SetHitsound extends Action<ActionType.SetHitsound> {
  gameSound: GameSound
  hitsound: HitSound
  hitsoundVolume: number
}

export interface SetGameSound extends Action<ActionType.SetGameSound> {
  gameSound: GameSound
  hitsound: HitSound
  volume: number
}

export interface PlaySound extends Action<ActionType.PlaySound> {
  hitsound: HitSound
  hitsoundVolume: number
  angleOffset: number
  eventTag: string
}

export interface RecolorTrack extends Action<ActionType.RecolorTrack> {
  startTile: Tile
  endTile: Tile
  gapLength: number
  trackColorType: TrackColorType
  trackColor: Color
  secondaryTrackColor: Color
  trackColorAnimDuration: number
  trackColorPulse: TrackColorPulse
  trackPulseLength: number
  trackStyle: TrackStyle
  trackGlowIntensity: number
  angleOffset: number
  eventTag: string
}

export interface MoveTrack extends Action<ActionType.MoveTrack> {
  startTile: Tile
  endTile: Tile
  gapLength: number
  duration: number
  positionOffset: Vector2
  rotationOffset: number
  scale: Vector2
  opacity: number
  angleOffset: number
  ease: Ease
  maxVfxOnly: ToggleBool
  eventTag: string
}

export interface SetFilter extends Action<ActionType.SetFilter> {
  filter: Filter
  enabled: ToggleBool
  intensity: number
  duration: number
  ease: Ease
  disableOthers: ToggleBool
  angleOffset: number
  eventTag: string
}

export interface HallOfMirrors extends Action<ActionType.HallOfMirrors> {
  enabled: ToggleBool
  angleOffset: number
  eventTag: string
}

export interface ShakeScreen extends Action<ActionType.ShakeScreen> {
  duration: number
  strength: number
  intensity: number
  fadeOut: ToggleBool
  angleOffset: number
  eventTag: string
}

export interface SetPlanetRotation
  extends Action<ActionType.SetPlanetRotation> {
  ease: Ease
  easeParts: number
  easePartBehavior: EasePartBehavior
}

export interface KillPlayer extends Action<ActionType.KillPlayer> {
  playAnimation: ToggleBool
  failMessage: string
  eventTag: string
}

export interface MoveDecorations extends Action<ActionType.MoveDecorations> {
  duration: number
  tag: string
  decorationImage: string
  positionOffset: Vector2
  pivotOffset: Vector2
  rotationOffset: number
  scale: Vector2
  color: Color
  opacity: number
  depth: number
  parallax: Vector2
  parallaxOffset: Vector2
  angleOffset: number
  ease: Ease
  eventTag: string
  maskingType: MaskingType
  useMaskingDepth: ToggleBool
  maskingFrontDepth: number
  maskingBackDepth: number
}

export interface PositionTrack extends Action<ActionType.PositionTrack> {
  positionOffset: Vector2
  relativeTo: Tile
  rotation: number
  scale: number
  opacity: number
  justThisTile: ToggleBool
  editorOnly: ToggleBool
  stickToFloors: ToggleBool
}

export interface Hold extends Action<ActionType.Hold> {
  duration: number
  distanceMultiplier: number
  landingAnimation: ToggleBool
}

export interface SetHoldSound extends Action<ActionType.SetHoldSound> {
  holdStartSound: HoldStartSound
  holdLoopSound: HoldLoopSound
  holdEndSound: HoldEndSound
  holdMidSound: HoldMidSound
  holdMidSoundType: HoldMidSoundType
  holdMidSoundDelay: number
  holdMidSoundTimingRelativeTo: HoldMidSoundTimingRelativeTo
  holdSoundVolume: number
}

export interface TileDimensions extends Action<ActionType.TileDimensions> {
  width: number
  length: number
}

export interface RepeatEvents extends Action<ActionType.RepeatEvents> {
  repetitions: number
  interval: number
  tag: string
}

export interface Bloom extends Action<ActionType.Bloom> {
  enabled: ToggleBool
  threshold: number
  intensity: number
  color: Color
  duration: number
  ease: Ease
  angleOffset: number
  eventTag: string
}

export interface SetConditionalEvents
  extends Action<ActionType.SetConditionalEvents> {
  perfectTag: string
  hitTag: string
  earlyPerfectTag: string
  latePerfectTag: string
  barelyTag: string
  veryEarlyTag: string
  veryLateTag: string
  missTag: string
  tooEarlyTag: string
  tooLateTag: string
  lossTag: string
}

export interface MultiPlanet extends Action<ActionType.MultiPlanet> {
  planets: PlanetCount
}

export interface FreeRoam extends Action<ActionType.FreeRoam> {
  duration: number
  size: Vector2
  positionOffset: Vector2
  outTime: number
  outEase: Ease
  hitsoundOnBeats: HitSound
  hitsoundOffBeats: HitSound
  countdownTicks: number
  angleCorrectionDir: number
}

export interface FreeRoamTwirl extends Action<ActionType.FreeRoamTwirl> {
  position: Vector2
}

export interface FreeRoamRemove extends Action<ActionType.FreeRoamRemove> {
  position: Vector2
  size: Vector2
}

export interface FreeRoamWarning extends Action<ActionType.FreeRoamWarning> {
  position: Vector2
}

export interface Pause extends Action<ActionType.Pause> {
  duration: number
  countdownTicks: number
  angleCorrectionDir: number
}

export interface AutoPlayTiles extends Action<ActionType.AutoPlayTiles> {
  enabled: ToggleBool
  safetyTiles: ToggleBool
}

export interface Hide extends Action<ActionType.Hide> {
  hideJudgment: ToggleBool
  hideTileIcon: ToggleBool
}

export interface ScaleMargin extends Action<ActionType.ScaleMargin> {
  scale: number
}

export interface ScaleRadius extends Action<ActionType.ScaleRadius> {
  scale: number
}

export interface ScreenTile extends Action<ActionType.ScreenTile> {
  tile: Vector2
  angleOffset: number
  eventTag: string
}

export interface ScreenScroll extends Action<ActionType.ScreenScroll> {
  scroll: Vector2
  angleOffset: number
  eventTag: string
}

export interface AddText extends Action<ActionType.AddText> {
  decText: string
  font: FontName
  position: Vector2
  relativeTo: DecPlacementType
  pivotOffset: Vector2
  rotation: number
  lockRotation: ToggleBool
  scale: Vector2
  lockScale: ToggleBool
  color: Color
  opacity: number
  depth: number
  parallax: Vector2
  tag: string
}

export interface SetText extends Action<ActionType.SetText> {
  decText: string
  tag: string
  angleOffset: number
  eventTag: string
}

export interface EditorComment extends Action<ActionType.EditorComment> {
  comment: string
}

export interface Bookmark extends Action<ActionType.Bookmark> {}

export interface CallMethod extends Action<ActionType.CallMethod> {
  method: string
  angleOffset: number
}

export interface AddComponent extends Action<ActionType.AddComponent> {
  component: string
  properties: string
  duration: number
  angleOffset: number
}

export interface ScalePlanets extends Action<ActionType.ScalePlanets> {
  duration: number
  targetPlanet: TargetPlanet
  scale: number
  angleOffset: number
  ease: Ease
  eventTag: string
}

export type LevelEvent =
  | SetSpeed
  | Twirl
  | Multitap
  | Checkpoint
  | CustomBackground
  | ChangeTrack
  | ColorTrack
  | AnimateTrack
  | Flash
  | MoveCamera
  | SetHitsound
  | SetGameSound
  | PlaySound
  | RecolorTrack
  | MoveTrack
  | SetFilter
  | HallOfMirrors
  | ShakeScreen
  | SetPlanetRotation
  | KillPlayer
  | MoveDecorations
  | PositionTrack
  | Hold
  | SetHoldSound
  | TileDimensions
  | RepeatEvents
  | Bloom
  | SetConditionalEvents
  | MultiPlanet
  | FreeRoam
  | FreeRoamTwirl
  | FreeRoamRemove
  | FreeRoamWarning
  | Pause
  | AutoPlayTiles
  | Hide
  | ScaleMargin
  | ScaleRadius
  | ScreenTile
  | ScreenScroll
  | SetText
  | EditorComment
  | Bookmark
  | CallMethod
  | AddComponent
  | ScalePlanets

export type DecorationEvent = AddDecoration | AddText
